/*
** Author: Dimitris Diochnos
** email : diochnos AT gmail DOT com
**
** File Name: mighty_classes.cpp
**
** Date Started : 20 April 2006.
** Date Finished: 28 April 2006.
**
** internals_mc (version 1.0)
** Date Started :  6 July 2007.
** Date Finished: 11 July 2007.
**
** internals_mc (version 1.1)
** Date Started : 12 July 2007.
** Date Finished: 20 July 2007.
**
** internals_mc (version 2.0)
** Date Started : 28 March 2008.
** Date Finished: 29 March 2008.
**
**
** Tab space: 3
**
**
** The "main" file.
**
*/

// So that you avoid warning for strcpy () in Visual Studio:
#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE 1
#endif

#ifdef _MSC_VER
#pragma comment(lib, "pthreadVC2.lib")
#endif

#include <fstream>
#include <iostream>
#include <iomanip>
#include <cstring>
#include <ctime>
#include <cctype>
//#include <unistd.h>			/* necessary for sysconf (_SC_NPROCESSORS_CONF) */
//#include <sys/time.h>		/* necessary for gettimeofday() -- should be deprecated by now ... */
#include <pthread.h>
#include "mighty_classes.h"
#include "sec_skills_mighty.h"
#include "queue.h"
#include "threads_defs.h"
//#include "math_toolbox.h"
#include "gettimeofday.h"	/* necessary for gettimeofday() */

#define CURRENT_VERSION_NUM 2.0
#define CURRENT_VERSION_STR "2.0"
#define CURRENT_VERSION_RELEASE_DATE "March 29, 2008"

/*
** External global variables.
*/
extern char * Alchemists_Names [];
extern char * Barbarians_Names [];
extern char * Beastmasters_Names [];
extern char * Death_Knights_Names [];
extern char * Demoniacs_Names [];
extern char * Knights_Names [];
extern char * Overlords_Names [];
extern char * Planeswalkers_Names [];
extern char * Rangers_Names [];

/*****************************************************************************/

unsigned char * Read_Selections (char *);
bool 	are_valid (HERO_PTR, unsigned char *, char *);
void default_version_message ();
void default_about_message ();
void default_error_message ();
void default_help_message ();
void default_log_message ();
void PrintSelections (unsigned char *);
int * generate_numbers (const int);
// The following two are deprecated.
//     R E M O V E ! ! ! ! ! !
//void * generator_function (void *);
//void * worker_function_1 (void *);
void * worker_function (void *);

/*****************************************************************************/

int 	main (int argc, char * argv [])
{
	//clock_t 			start, finish;	// This is no longer used.
	unsigned char 	pointer;
	int				i, j; // counters. re-use them if you want at any part of the code.
	long unsigned int episodes, episodes_array [WORKER_THREADS];
	long unsigned int temp_value;
	int 					algorithm;
	unsigned char * selections;
	DecisionMaker * dMaker [WORKER_THREADS];
	struct timeval start, end;
	TIME_DIFF *	difference;

	if (argc == 2) {
		char * request = new char [strlen(argv[1]) + 1];

		strcpy(request, argv[1]);
		for (pointer = 0; pointer < strlen (argv [1]); pointer++)
			request [pointer] = toupper ( request [pointer] );

		if (strcmp(request, "-VERSION") == 0)
			default_version_message ();
		else if (strcmp(request, "-HELP") == 0)
			default_help_message ();
		else if (strcmp(request, "-ABOUT") == 0)
			default_about_message ();
		else if (strcmp(request, "-LOG") == 0)
			default_log_message ();
		else {
			default_error_message ();

			exit(1);
		}

		delete [] request;

		exit(0);
	}
	if ((argc != 5) && (argc != 6)) {
		default_error_message ();

		exit (1);
	}


	/*
	** Program has 4 or 5 arguments. Prepare for execution.
	*/


	char * 	given_class = new char [strlen (argv [1]) + 1];
	char * 	given_name = new char [strlen (argv [2]) + 1];

	// Convert to UPPERCASE.
	strcpy (given_class, argv [1]);
	for (pointer = 0; pointer < strlen (argv [1]); pointer++)
		given_class [pointer] = toupper ( given_class [pointer] );

	// Convert to UPPERCASE.
	strcpy (given_name, argv [2]);
	for (pointer = 0; pointer < strlen (argv [2]); pointer++)
		given_name [pointer] = toupper ( given_name [pointer] );

	/*
	** Validate user input.
	*/
	unsigned char 	classification;
	unsigned char 	name;

	if (strcmp (given_class, "ALCHEMIST") == 0 )
		classification = ALCHEMIST;
	else if (strcmp (given_class, "BARBARIAN") == 0 )
		classification = BARBARIAN;
	else if (strcmp (given_class, "BEASTMASTER") == 0 )
		classification = BEASTMASTER;
	else if (strcmp (given_class, "DEATH_KNIGHT") == 0 )
		classification = DEATH_KNIGHT;
	else if (strcmp (given_class, "DEMONIAC") == 0 )
		classification = DEMONIAC;
	else if (strcmp (given_class, "KNIGHT") == 0 )
		classification = KNIGHT;
	else if (strcmp (given_class, "OVERLORD") == 0 )
		classification = OVERLORD;
	else if (strcmp (given_class, "PLANESWALKER") == 0 )
		classification = PLANESWALKER;
	else if (strcmp (given_class, "RANGER") == 0 )
		classification = RANGER;
	else {
		// Free-up memory for first two arguments.
		delete 	[] given_class;
		delete 	[] given_name;

		std::cerr << "Invalid Hero class given." << std::endl;
		std::cerr << "Please execute program with -help parameter to view help." << std::endl;

		exit (2);
	}

	switch(classification) {
		case ALCHEMIST:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Alchemists_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case BARBARIAN:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Barbarians_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case BEASTMASTER:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Beastmasters_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case DEATH_KNIGHT:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Death_Knights_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case DEMONIAC:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Demoniacs_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case KNIGHT:
			for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) {
				if (strcmp (given_name, Knights_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case OVERLORD:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Overlords_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case PLANESWALKER:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Planeswalkers_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case RANGER:
			for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) {
				if (strcmp (given_name, Rangers_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		default:
name_error:
			std::cerr << "Invalid Hero name given." << std::endl;
			std::cerr << "Please execute program with -help parameter to view help." << std::endl;

			// Free-up memory for first two arguments.
			delete 	[] given_class;
			delete 	[] given_name;

			exit (4);

			break;
	}


	/*
	** Procceed to execution of Always Left policy.
	*/
normal_execution:

	// Free-up memory for first two arguments.
	delete 	[] given_class;
	delete 	[] given_name;

	char * 	given_algorithm = new char [strlen (argv [3]) + 1];

	strcpy (given_algorithm, argv [3]);
	for (pointer = 0; pointer < strlen (argv [3]); pointer++)
		given_algorithm [pointer] = toupper ( given_algorithm [pointer] );

	if (strcmp (given_algorithm, "AR") == 0 )
		algorithm = AR_ALGO;
	else if (strcmp (given_algorithm, "ANSA") == 0 )
		algorithm = ANSA_ALGO;
	else if (strcmp (given_algorithm, "AL") == 0 )
		algorithm = AL_ALGO;
	else if (strcmp (given_algorithm, "ALTP") == 0 )
		algorithm = ALTP_ALGO;
	else if (strcmp (given_algorithm, "SPOU") == 0 )
		algorithm = SPOU_ALGO;
	//else if (strcmp (given_algorithm, "FMTP") == 0 )
	//	algorithm = FMTP_ALGO;
	else {
		delete 	[] given_algorithm;

		std::cerr << "Invalid algorithm given." << std::endl;
		std::cerr << "Please execute program with -help parameter to view help." << std::endl;

		exit (5);
	}

	// Free-up memory.
	delete 	[] given_algorithm;



	// Create an instance of the selected hero for each thread.
	//HERO_PTR	yet_another_hero = new HERO (classification, name);
	HERO_PTR	yet_another_hero [WORKER_THREADS];
	HERO_PTR	common_hero = new HERO (classification, name);
	for (i = 0; i < WORKER_THREADS; i++)
		yet_another_hero [i] = new HERO (classification, name);



	// Initialize DecisionMaker based on algorithm selection.
	switch(algorithm) {
		//case AR_ALGO:
		case ANSA_ALGO:
			for (i = 0; i < WORKER_THREADS; i++)
				dMaker [i] = new HeroAlwaysRight (yet_another_hero [i]);
			break;

		case AL_ALGO:
			for (i = 0; i < WORKER_THREADS; i++)
				dMaker [i] = new HeroAlwaysLeft (yet_another_hero [i]);
			break;

		case ALTP_ALGO:
			selections = Read_Selections ( DEFAULT_ALTP_PREFS_FILE );
			// Same initialization for all instances ==> It suffices to check one of them (--> index: 0).
			if (are_valid (yet_another_hero [0], selections, DEFAULT_ALTP_PREFS_FILE))
				for (i = 0; i < WORKER_THREADS; i++)
					dMaker [i] = new HeroAlwaysLeftThenPref (yet_another_hero [i], selections);
			else {
				// Problem with the entries in the file.
				// Deallocate memory and return.
				for (i = 0; i < WORKER_THREADS; i++)
					delete yet_another_hero [i];
				delete	common_hero;
				delete 	[] selections;
				exit(6);
			}
			delete 	[] selections;
			break;

		case SPOU_ALGO:
			selections = Read_Selections ( DEFAULT_SPOU_PREFS_FILE );
			//std::cout << "about to print selections" << std::endl;
			//PrintSelections (selections);
			//std::cout << "selections printed" << std::endl;
			//
			// Similar Case as above:
			// Same initialization for all instances ==> It suffices to check one of them.
			if (are_valid (yet_another_hero [0], selections, DEFAULT_SPOU_PREFS_FILE)) {
				for (i = 0; i < WORKER_THREADS; i++)
					dMaker [i] = new HeroSeekPrefOtherwiseUpgrade (yet_another_hero [i], selections);
				//dynamic_cast<HeroSeekPrefOtherwiseUpgrade *>(dMaker)->printPreferences ();
			}
			else {
				// Problem with the entries in the file.
				// Deallocate memory and return.
				for (i = 0; i < WORKER_THREADS; i++)
					delete yet_another_hero [i];
				delete	common_hero;
				delete 	[] selections;
				exit(6);
			}
			delete 	[] selections;
			break;

//		case FMTP_ALGO:
//			std::cerr << "The selected algorithm is not supported yet!" << std::endl;
//			exit(7);
			//dMaker = new HeroAlwaysLeft (yet_another_hero);
//			break;

		default:
			std::cerr << "You should not enter here while selecting policy!" << std::endl;
			exit(7);
			break;
	}



	/*
	** Fetch number of episodes.
	*/
	episodes = static_cast<long unsigned int> ( atoi( argv[4] ) );
	for (i = 0; i < WORKER_THREADS; i++)
		episodes_array [i] = static_cast<long unsigned int> (episodes / WORKER_THREADS);
	// The first thread should compute the remaining episodes as well.
	// Silly attempt to hide latency on thread creation.
	episodes_array [0] += static_cast<long unsigned int> (episodes % WORKER_THREADS);




	/*
	** Fetch printing procedure.
	** By default use FANCY output.
	*/
	unsigned char	output_algo = FANCY;
	if (argc == 6) {
		char * requested_output = new char [strlen(argv[5]) + 1];
		strcpy (requested_output, argv[5]);
		for (pointer = 0; pointer < strlen(argv[5]) + 1; pointer++)
			requested_output[pointer] = toupper( requested_output[pointer] );		// Convert to uppercase.

		if (strcmp(requested_output, "DEPRECATED") == 0)
			output_algo = DEPRECATED;
		else if (strcmp(requested_output, "NEAT") == 0)
			output_algo = NEAT;
		else if (strcmp(requested_output, "FANCY") != 0)
			std::cerr << "Warning: Invalid output requested. FANCY used instead." << std::endl;

		delete	[] requested_output;
	}


	// Determine the seed for the randomized process.
	unsigned int seed;
	// The following is uncommented while debugging.
	//seed = 2008;	// current year.
	// Make it "truly random".
	seed = (unsigned int) time ( NULL );
	srand( seed );
	//std::cout << "Seed        : " << seed << std::endl;


	// Store user preferences for number of episodes and printing method.
	// Also store seed just in case it is needed for debugging within the
	// Math Toolbox.
	MT_PTR myMathToolbox [WORKER_THREADS];
	for (i = 0; i < WORKER_THREADS; i++)
		myMathToolbox [i] = new MT;

	for (i = 0; i < WORKER_THREADS; i++) {
		myMathToolbox [i]->setNumberOfEpisodes ( episodes_array [i] );
		if (output_algo != FANCY)
			myMathToolbox [i]->setPrintingMethod ( output_algo );
		myMathToolbox [i]->setSeed ( seed );
	}


	/*
	** Apply MonteCarlo with selected options.
	*/
	MONTE_CARLO_PTR MC [WORKER_THREADS];
	for (i = 0; i < WORKER_THREADS; i++)
		MC [i] = new MonteCarlo ( yet_another_hero [i], dMaker [i] );




	/*********************************************\
	* Create necessary structures for threads and *
	* pass necessary objects to them!             *
	\*********************************************/
	// create opaque objects.
	pthread_t * threads = new pthread_t [NUM_THREADS];
	// Make threads Joinable for sure.
	pthread_attr_t attr;
	pthread_attr_init (&attr);
	pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_JOINABLE);
	// Declare queue mutex and associated condition variable for each worker thread.
	pthread_mutex_t ** q_mutex_ptr_array = new PTHREAD_MUTEX_T_PTR [WORKER_THREADS];
	pthread_cond_t ** q_cond_ptr_array = new PTHREAD_COND_T_PTR [WORKER_THREADS];
	// Declare queue and associated waiting variable for each worker thread.
	QUEUE_PTR * q_ptr_array = new QUEUE_PTR [WORKER_THREADS];
	bool ** waiting = new BOOL_PTR [WORKER_THREADS];
	//bool gen_finished = false;	// Indicates whether the generator finished its job.
	// Create one queue for each thread.
	for (i = 0; i < WORKER_THREADS; i++) {
		q_ptr_array [i] = new QUEUE;	// Point to a new queue.
		q_mutex_ptr_array [i] = new pthread_mutex_t;
		q_cond_ptr_array [i] = new pthread_cond_t;
		waiting [i] = new bool;
		*(waiting [i]) = false;		// Nobody is waiting initially.
	}
	// Initialize queue mutex and condition variable for each worker-thread.
	for (i = 0; i < WORKER_THREADS; i++) {
		pthread_mutex_init (q_mutex_ptr_array [i], NULL);
		pthread_cond_init (q_cond_ptr_array [i], NULL);
	}
	// variable indicating the amount of work performed at any given time by *all* worker threads.
	long unsigned int		total_episodes_generated_atm = 0;
	// Declare and initialize appropriate mutex for 'total_episodes' variable.
	pthread_mutex_t * episodes_mutex = new pthread_mutex_t;
	pthread_mutex_init (episodes_mutex, NULL);	
	// Initialize Data Structures for various threads.
	WORKER_DATA * data_array = new WORKER_DATA [WORKER_THREADS];
	GENERATOR_DATA gen_data;
	gen_data.queue_mutex_array = q_mutex_ptr_array;
	gen_data.queue_cond_array = q_cond_ptr_array;
	gen_data.queue_array = q_ptr_array;
	gen_data.id = GENERATOR_ID;
	gen_data.items_per_chunk = RND_NUMS_SIZE;
	//gen_data.worker_threads = WORKER_THREADS;
	//gen_data.episodes_requested = episodes;
	//gen_data.floor_episodes_per_thread = episodes / WORKER_THREADS;
	gen_data.waiting_array = waiting;
	for (i = 0; i < WORKER_THREADS; i++) {
		data_array [i].queue_mutex = q_mutex_ptr_array [i];
		data_array [i].queue_cond = q_cond_ptr_array [i];
		data_array [i].queue = q_ptr_array [i];
		data_array [i].id = i;	// Note that worker threads should get ids \in {0,1,2,...,WORKER_THREADS-1}.
		data_array [i].total_episodes_computed_by_threads = episodes;
		data_array [i].episodes_to_compute = episodes_array [i];
		data_array [i].total_episodes_atm = &total_episodes_generated_atm;
		data_array [i].total_episodes_mutex = episodes_mutex;
		data_array [i].waiting = waiting [i];
		data_array [i].myMC = MC [i];
		data_array [i].myMT = myMathToolbox [i];
	}


	/*
	** All initializations took place!
	** Now create threads perform MULTI-THREADED computations! :-)
	*/
	// Count time!
	gettimeofday (&start, NULL);
	//int const number_of_cores = sysconf (_SC_NPROCESSORS_CONF);
	int rc;	// to store return code by pthread_create ().
	for (i = 0; i < WORKER_THREADS; i++) {
		rc = pthread_create(&(threads [i]), &attr, worker_function, (void *) &(data_array[i]));
		if (rc) {
			std::cerr << "ERROR; return code from pthread_create() is " << rc << std::endl;
			exit (8);
		}
	}
	int const bound = episodes / WORKER_THREADS;
	for (i = 0; i < WORKER_THREADS; i++) {
		/* Lock respective queue-mutex. */
		pthread_mutex_lock (gen_data.queue_mutex_array [i]);
		for (j = 0; j < bound; j++)
			gen_data.queue_array [i]->enqueue ( generate_numbers (gen_data.items_per_chunk) );
		if (i == 0) {
			// Generate (if necessary) remainder number of episodes for the first thread because they are assigned there.
			for (j = 0; j < (int) (episodes % WORKER_THREADS); j++)
				gen_data.queue_array [0]->enqueue ( generate_numbers (gen_data.items_per_chunk) );
		}
		// Trigger a new thread. A core is "waiting". :-)
		if (gen_data.waiting_array [i])
			pthread_cond_signal (gen_data.queue_cond_array [i]);
		/* unlock appropriate queue-mutex. */
		pthread_mutex_unlock (gen_data.queue_mutex_array [i]);
	}

	/* Wait for all threads to join. */
	void * status;
	for (i = 0; i < WORKER_THREADS; i++)
		pthread_join ( threads [i], &status );
	gettimeofday (&end, NULL);


	// Clear screen from counter.
	std::cout << "\r                            \r";
	std::cout.flush ();


	/*
	** Now merge partial results and compute actual running time.
	*/
	MT_PTR overall_results = new MT;
	overall_results->initializeOccurrences ();	// Not needed.
	overall_results->setNumberOfEpisodes ( episodes );
	overall_results->setPrintingMethod ( data_array [0].myMT->getPrintingMethod () );
	overall_results->setSeed ( data_array [0].myMT->getSeed () );
	for (i = 0; i < DIFF_SEC_SKILLS; i++) {
		for (j = 0; j < WORKER_THREADS; j++) {
			temp_value = overall_results->getOccurrences ((unsigned char) i);
			overall_results->setOccurrences ((unsigned char) i, temp_value + data_array [j].myMT->getOccurrences((unsigned char) i));
		}
	}
	difference = my_difftime (&start, &end);
	
	/*
	** Print Results for specific hero.
	*/

	if ((algorithm == ANSA_ALGO) || (algorithm == AL_ALGO)) {
		if (episodes >= 400000000) {
			std::cerr << "\aSmoothing results omitted due to possible overflow error." << std::endl;
			goto print_results;
		}
		overall_results->smoothResults (common_hero);	// Smooth Results if necessary.
	}
print_results:
	//overall_results->printStatistics_THREADS (common_hero); 
	overall_results->printStatistics_THREADS (common_hero, difference);



	/*
	** Deallocate memory reserved earlier.
	**
	** Note, that the process of deallocation is
	** performed roughly in the order it was allocated.
	*/
	for (i = 0; i < WORKER_THREADS; i++) {
		delete yet_another_hero [i];
		delete dMaker [i];
		delete myMathToolbox [i];
		delete MC [i];
	}
	delete	common_hero;


	/*
	**
	** Thread-related memory.
	**
	*/

	// Just in case a future release changes, I am still using the
	// definition of NUM_THREADS. So deallocate this time like:
	//for (i = 0; i < NUM_THREADS; i++)
	//	delete threads [i];
	delete [] threads;

	for (i = 0; i < WORKER_THREADS; i++) {
		/* QUEUES */
		/* 
		** Set a destroy-QUEUE () function and call this one before you exit
		** Just to be cautious.
		*/
		delete q_ptr_array [i];

		/* MUTECES */
		pthread_mutex_destroy (q_mutex_ptr_array [i]);
		delete q_mutex_ptr_array [i];

		/* CONDITION VARIABLES */
		pthread_cond_destroy(q_cond_ptr_array [i]);
		delete q_cond_ptr_array [i];

		/* BOOLEAN WAITING ARRAY */
		delete waiting [i];
	}
	/* DELETE arrays of pointers for QUEUE muteces and condition variables. */
	delete [] q_mutex_ptr_array;
	delete [] q_cond_ptr_array;
	/* Also, delete arrays of pointers for QUEUES and BOOLEAN waiting vars. */
	delete [] q_ptr_array;
	delete [] waiting;
	/* Delete episodes_mutex */
	pthread_mutex_destroy (episodes_mutex);
	delete episodes_mutex;
	/* Finally, delete array storing DATA for various threads. */
	delete [] data_array;


	/* Delete memory reserved for Math Toolbox used for cumulative results. */
	delete overall_results;

	/* Delete data structure storing time difference. */
	delete difference;



	/*
	** The following deallocations are old and unnecessary.
	** If you are not going to revert back to a single core version,
	** then probably it's best to remove them from here.
	** ELIMINATE MESS!!! :)
	*/
	// Free-up memory reserved.
	// This needs to be fixed.
	/*delete	yet_another_hero;
	delete	dMaker;
	delete	MC;
	delete	myMT;*/

	//char tempoulas[32];
	//std::cin.getline (tempoulas, 31);


	return 	0;
}

/*****************************************************************************/

	/*
	** From now on (version 2.0) only this function is allowed
	** to make calls to the random number generator.
	** It creates a valid sequence of remainders that
	** can be used by all subsequent 'dice-rolls'.
	**
	** IMPORTANT: DIRECTLY USE CALLS ON rand() FUNCTION.
	** FOR THE MOD OPERATOR HOLDS:
	** (9 % 3) % 2 =!= (9 % 2) % 3
	**
	** ============> DO NOT TAKE MOD 'TOTAL_CHANCES' = 112 :)
	** THIS WILL ''NOT'' GIVE YOU A SEQUENCE WITH 'GOOD' STATISTICAL PROPERTIES.
	*/
int * generate_numbers (const int size)
{
	int * numbers = new int [size];

	for (int i = 0; i < size; i++)
		numbers [i] = rand ();

	return numbers;
}

/*****************************************************************************/

	/**********************************************\
	* Functions that threads use in order to work! *
	\**********************************************/

void *	worker_function (void * input)
{
	register long unsigned int episodes_generated = 0;
	//int i;
	int * numbers;
	WORKER_DATA * data_struct = (WORKER_DATA *) input;
	long unsigned int const episodes_to_generate = data_struct->episodes_to_compute;
	long unsigned int const episodes_by_all = data_struct->total_episodes_computed_by_threads;
	//int const array_size = data_struct->items_per_chunk;
	const bool show_progress = (episodes_by_all >= WORKER_THREADS * EPISODES_FLUSH_CONSTANT) ? true : false;

	// Refer to MonteCarlo::run () for more info about the following function.
	data_struct->myMT->initializeOccurrences ();

	/* Lock queue-mutex. */
	pthread_mutex_lock (data_struct->queue_mutex);
	if (((data_struct->queue)->get_size () == 0) && (episodes_to_generate > 0) ) {
		// Indicate that somebody is waiting!
		*(data_struct->waiting) = true;
		pthread_cond_wait (data_struct->queue_cond, data_struct->queue_mutex);
		*(data_struct->waiting) = false;
	}
	/* unlock queue-mutex - it won't make any difference though... */
	pthread_mutex_unlock (data_struct->queue_mutex);

	while (episodes_generated < episodes_to_generate) {		
		numbers = (data_struct->queue)->dequeue ();
		// The following is for debugging purposes.
		//if (numbers == (int *) NULL)
		//	std::cerr << "   !!!!!! NULL pointer returned !!!!!! [1, " << data_struct->id << "]\a" << std::endl;
		
		/* Generate an episode based on your random sequence. */
		data_struct->myMC->run_once ( data_struct->myMT, numbers );
		
		delete [] numbers;
		episodes_generated++;

		// Update the "global" counter every 100 episodes (or so) to avoid overhead by muteces.
		if (episodes_generated % EPISODES_FLUSH_CONSTANT == 0) {
			// Lock total episodes mutex.
			pthread_mutex_lock (data_struct->total_episodes_mutex);
			*(data_struct->total_episodes_atm) += EPISODES_FLUSH_CONSTANT;
			if ( show_progress /*&& (*(data_struct->total_episodes_atm) % ((int) episodes_by_all / 100)) == 0*/ ) {
				std::cout << "\rWork done    : " << (int) (*(data_struct->total_episodes_atm)) *100 / episodes_by_all << " %";
				std::cout.flush ();
			}
			// Unlock mutex so that others can report their progress.
			pthread_mutex_unlock (data_struct->total_episodes_mutex);
		}
	}
	
	// Make a last report with remaining episodes (if any).
	// Lock total episodes mutex.
	pthread_mutex_lock (data_struct->total_episodes_mutex);
	*(data_struct->total_episodes_atm) += episodes_generated % EPISODES_FLUSH_CONSTANT;
	if ( show_progress /*&& (*(data_struct->total_episodes_atm) % ((int) episodes_by_all / 100)) == 0*/) {
		std::cout << "\rWork done    : " << (int) (*(data_struct->total_episodes_atm)) *100 / episodes_by_all << " %";
		std::cout.flush ();
	}
	// Unlock mutex so that others can report their progress.
	pthread_mutex_unlock (data_struct->total_episodes_mutex);

	//std::cout << " returning " << data_struct->id << std::endl;

	/* exit - join () - return results. */
	pthread_exit (NULL);

	return NULL;
}


/*****************************************************************************/

unsigned char * Read_Selections (char * name)
{
	int i, j;
	char current_line [14];
	bool already_found[DIFF_SEC_SKILLS];
	unsigned char * selections = (unsigned char *) NULL;

	/* Check if file 'selections.txt' exists. */
	std::ifstream selFD ( name );

	if ( !selFD ) {
		std::cerr << "Usage: Program expects file " << name << " in same directory.\n";

		exit (6);
	}

	// Initialize "double_appearence" array.
	for (i = 0; i < DIFF_SEC_SKILLS; i++)
		already_found[i] = false;	// No skill is already read.

	/*
	** Store user's selections in memory.
	*/
	selections = new unsigned char [DIFF_SEC_SKILLS];
	
	for (i = 0; i < (DIFF_SEC_SKILLS - 1); i++) {
		selFD.getline (current_line, 13);
		for (j = 0; j < (int) strlen (current_line ); j++) //convert selection to uppercase
				current_line[j] = (char) toupper( (int) current_line[j] );
		for (j = 1; j < DIFF_SEC_SKILLS; j++) {
			if ( strcmp( current_line, get_skill_name_consecutive (j) ) == 0 ) {
				if ( !already_found[j] )
					already_found [j] = true;
				else {
					// We have a problem because same skill was read earlier in file.
					// Indicate error to user.
					std::cerr << name << ": Duplicate skill (" << get_skill_name_consecutive (j);
					std::cerr << ") on input line " << (int) (i + 1) << "." << std::endl;

					exit(6);
				}
				selections[i] = j; 	// Store preference.
				break;
			}
		}
		if ( j == DIFF_SEC_SKILLS ) {
			// Parser was unable to locate skill given by user.
			std::cerr << name << ": Invalid skill given on input line " << (int) (i + 1) << "." << std::endl;

			exit (6);
		}
		//std::cout << get_skill_name_consecutive(k) << "." << std::endl;
	}
	selections[DIFF_SEC_SKILLS - 1] = NULL_SKILL; 	// Null-skill no-one wants it.

	selFD.close ();

	return selections;
}

/*****************************************************************************/

bool 	are_valid (HERO_PTR yet_another_hero, unsigned char * selections, char * filename)
{
	int i;
	bool found;
	//char temp[32];

	found = false;
	//std::cout << "selections[0] : " << get_skill_name (selections[0]) << std::endl;
	//std::cout << "slot      [0] : " << get_skill_name (yet_another_hero->get_skill_from_slot (0)) << std::endl;
	for (i = 0; i < SKILL_SLOTS; i++) {
		if (selections[i] == yet_another_hero->get_skill_from_slot (0)) {
			found = true;
			break;
		}
	}
	if (!found) {
		std::cerr << "Hero starts with " << get_skill_name( yet_another_hero->get_skill_from_slot (0) );
		std::cerr << " and you did not provide it in the first 8 skills." << std::endl;
		std::cerr << "Please edit your selections in " << filename << " file." << std::endl;

		//std::cin.getline ( temp, 31 );

		return false;;
	}

	if (yet_another_hero->get_skill_from_slot(1) != NULL_SKILL) {
		found = false;
		for (i = 0; i < SKILL_SLOTS; i++) {
			if (selections[i] == yet_another_hero->get_skill_from_slot (1)) {
				found = true;
				break;
			}
		}

		if (!found)	{
			std::cerr << "Hero starts with " << get_skill_name( yet_another_hero->get_skill_from_slot (1) );
			std::cerr << " and you did not provide it in the first 8 skills." << std::endl;
			std::cerr << "Please edit your selections in " << filename << " file." << std::endl;

			//std::cin.getline ( temp, 31 );

			return false;
		}
	}

	return true;
}

/*****************************************************************************/

void default_version_message ()
{
	std::cout << CURRENT_VERSION_STR << std::endl;
}

void default_about_message ()
{
	std::cout << "internals_mc: about" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~~~" << std::endl;
	std::cout << "Version: " << CURRENT_VERSION_STR << std::endl;
	std::cout << "Release: " << CURRENT_VERSION_RELEASE_DATE << std::endl;
	std::cout << "* For a complete documentation and comments please visit homepage:" << std::endl;
	std::cout << "  http://heroescommunity.com/viewthread.php3?TID=17812&pagenumber=8" << std::endl;
	std::cout << "* Bugs and bugfixes can be reported in the forum above (recommended)" << std::endl;
	std::cout << "  or with a direct mail at: diochnos AT gmail DOT com." << std::endl;
}

void default_error_message ()
{
	std::cout << "internals_mc" << std::endl;
	std::cout << "~~~~~~~~~~~~" << std::endl;
	std::cout << "Incorrect parameters. Please execute:" << std::endl;
	std::cout << "internals_mc -help" << std::endl;
	std::cout << "for a complete list of available options or visit homepage at:" << std::endl;
	std::cout << "http://heroescommunity.com/viewthread.php3?TID=17812&pagenumber=8" << std::endl;
}

void default_help_message ()
{
	int pointer;

	std::cout << "internals_mc: help" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~~" << std::endl;
	std::cout << "Usage: Program expects 4 or 5 arguments." << std::endl;
	std::cout << "1st  : Hero Class." << std::endl;
	std::cout << "2nd  : Hero Name." << std::endl;
	std::cout << "3rd  : Policy (Skill Selection Algorithm)." << std::endl;
	std::cout << "4th  : Number of episodes." << std::endl;
	std::cout << "5th  : Printing method (optional argument)." << std::endl << std::endl;
	std::cout << "Valid names for Hero Classes:" << std::endl;
	std::cout << "Alchemist, Barbarian, Beastmaster, Death_Knight, " << std::endl;
	std::cout << "Demoniac, Knight, Overlord, Planeswalker, Ranger." << std::endl << std::endl;
	std::cout << "Valid names for Heroes:" << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Alchemists_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Barbarians_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Beastmasters_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Death_Knights_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Demoniacs_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) { 	// There is also Sir Mullich.
		std::cout << Knights_Names [pointer] << " ";
		if (pointer == 4)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Overlords_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Planeswalkers_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) { 	// There is also Gelu.
		std::cout << Rangers_Names [pointer] << " ";
		if (pointer == 4)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl << std::endl;
	std::cout << "Valid names for algorithms (policies):" << std::endl;
	std::cout << "AR  : Always Right." << std::endl;
	std::cout << "ANSA: Always Right (same as above)." << std::endl;
	std::cout << "AL  : Always Left." << std::endl;
	std::cout << "ALTP: Always Left-Then Preference." << std::endl;
	std::cout << "SPOU: Seek Preference-Otherwise Upgrade (according to Preference)." << std::endl;
	//std::cout << "FMTP: First Magic-Then Preference" << std::endl;
	std::cout << std::endl << std::endl;

	std::cout << "Valid names for Printing Methods:" << std::endl;
	std::cout << "DEPRECATED: The initial (raw) output." << std::endl;
	std::cout << "            Might be useful in conjuction with 'ansa2excel' program." << std::endl;
	std::cout << "NEAT      : A neat version for the deprecated output." << std::endl;
	std::cout << "FANCY     : Generates FANCY output." << std::endl;
	std::cout << "            This is used by default." << std::endl;

	std::cout << std::endl;
}

void	default_log_message ()
{
	std::cout << "internals_mc: log" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~" << std::endl << std::endl;
	std::cout << "Version 1.0:" << std::endl;
	std::cout << " * The initial version of the program." << std::endl;
	std::cout << " - Developed: July 6 - 11, 2007." << std::endl;
	std::cout << " - Released : July 11, 2007." << std::endl;
	std::cout << std::endl;
	std::cout << "Version 1.1:" << std::endl;
	std::cout << " * Fixed bugs when reading preference files for ALTP and SPOU policies." << std::endl;
	std::cout << " * Optimization enhancements." << std::endl;
	std::cout << " * Generation of a Mathematical toolbox that allows:" << std::endl;
	std::cout << "   > Smoothed results for AR and AL policies." << std::endl;
	std::cout << "   > Different printing methods of the results." << std::endl;
	std::cout << " * Fixed bug in SPOU policy behaviour. More info on thread." << std::endl;
	std::cout << " * Allowed the use of this log." << std::endl;
	std::cout << " - Developed: July 12 - 20, 2007." << std::endl;
	std::cout << " - Released : July 20, 2007." << std::endl;
	std::cout << "Version 2.0:" << std::endl;
	std::cout << " * POSIX threads are implemented in order to speed up computations." << std::endl;
	std::cout << " - Developed: March 28- April 7, 2008." << std::endl;
	std::cout << " - Released : " << __DATE__ << ", " << __TIME__ << "." << std::endl;
}

void 	PrintSelections(unsigned char * selections)
{
	int i;

	std::cout << "SELECTIONS" << std::endl;
	for (i = 0; i < DIFF_SEC_SKILLS; i++) {
		std::cout << get_skill_name_12 ( i ) << ": " << (int) selections[i] << std::endl;
	}
	std::cout << "----------" << std::endl;
}
