/*
** Author: Dimitris Diochnos
** email : diochnos AT gmail DOT com
**
** File Name: mighty_classes.cpp
**
** Date Started : 20 April 2006.
** Date Finished: 28 April 2006.
**
** internals_mc (version 1.0)
** Date Started :  6 July 2007.
** Date Finished: 11 July 2007.
**
** internals_mc (version 1.1)
** Date Started : 12 July 2007.
** Date Finished: 20 July 2007.
**
**
** Tab space: 3
**
**
** Definitions for the 'Hero' class.
**
*/

#ifndef __MIGHTY_CLASSES_H__
#define __MIGHTY_CLASSES_H__

#include <ctime>
#include "sec_skills_mighty.h"
/*
** Only the math_toolbox uses gettimeofday().
** So in a future release this should move to the header of the math toolbox.
*/
#include "gettimeofday.h"
//#include "math_toolbox.h" // Make it consistent at a later version.

/*****************************************************************************/

struct skill_offer{
	unsigned char left;
	unsigned char right;
};

typedef struct skill_offer SKILL_OFFER;
typedef SKILL_OFFER * SKILL_OFFER_PTR;

/*****************************************************************************/

class Hero {
	public:
								Hero (const unsigned char, const unsigned char name);
								~Hero ();
		void 					myInit (const unsigned char, const unsigned char name);
		unsigned char     get_magic_schools () const;
		unsigned char 		get_max_number_of_magic_schools () const;
		bool              get_has_wisdom () const;
		bool 					get_fully_upgraded () const;
		unsigned char     get_hero_level () const;
		unsigned char 		get_last_magic_level_offer () const;
		unsigned char 		get_last_wisdom_level_offer () const;
		unsigned char 		get_hero_class () const;
		unsigned char 		get_hero_name () const;
		unsigned char 		get_skill_from_slot (const unsigned char) const;
		unsigned char 		get_skill_slot (const unsigned char) const;
		unsigned char		get_initial_skills () const;
		unsigned char *	get_skill_chances () const;
		SKILL_OFFER_PTR 	get_skill_offer ();
		SKILL_OFFER_PTR 	get_skill_offer_THREADS (const int *, const int);	// Added with version 2.0
		void 					increment_level ();
		double *          ANSA_policy ();
		void              print_Name () const;
		void 					addSkill (const unsigned char);
		bool 					is_magic_school (const unsigned char) const;


	private:
		void 					depth_first_search (const unsigned char, bool *, unsigned char *, const unsigned char, bool, const double, const unsigned char, double *);
		unsigned char 		Roll_Dice_for_new_skill (const unsigned char) const;
		unsigned char 		Roll_Dice_for_new_magic_skill (const unsigned char) const;
		unsigned char 		Roll_Dice_for_magic_skill_upgrade (unsigned char [], const unsigned char) const;
		unsigned char 		Roll_Dice_for_skill_upgrade (unsigned char [], const unsigned char, const unsigned char) const;
		// Duplicates of the above that co-operate with THREADS. - Added with version 2.0
		unsigned char 		Roll_Dice_for_new_skill_THREADS (const unsigned char, const int *, int) const;
		unsigned char 		Roll_Dice_for_new_magic_skill_THREADS (const unsigned char, const int *, int) const;
		unsigned char 		Roll_Dice_for_magic_skill_upgrade_THREADS (unsigned char [], const unsigned char, const int *, int) const;
		unsigned char 		Roll_Dice_for_skill_upgrade_THREADS (unsigned char [], const unsigned char, const unsigned char, const int *, int) const;

		unsigned char     classification;
		unsigned char     name;
		unsigned char     level;
		unsigned char     initial_skills;
		unsigned char     skills [SKILL_SLOTS];
		unsigned char     expertize [SKILL_SLOTS];
		unsigned char     magic_schools;
		unsigned char 		max_number_of_magic_schools;
		bool              has_wisdom;
		unsigned char *   skill_chances;
		unsigned char 		last_magic_level_offer;
		unsigned char 		last_wisdom_level_offer;
		bool 					available_skills [DIFF_SEC_SKILLS];
		unsigned char 		mapping_skills [DIFF_SEC_SKILLS];
		unsigned char 		free_slot;
};

typedef Hero HERO;
typedef HERO * HERO_PTR;

/*****************************************************************************/

char * 	get_skill_name_12 (const unsigned char pointer);
char * 	get_skill_name (const unsigned char pointer);
char * 	get_skill_name_consecutive (const unsigned char pointer);

/*****************************************************************************/

class DecisionMaker {
	public:
										DecisionMaker (HERO_PTR);
										~DecisionMaker ();
		virtual unsigned char 	applyPolicy (SKILL_OFFER_PTR) = 0;
		virtual void				LoopInitializer () = 0;

	protected:
		HERO_PTR						myHero;
};

typedef DecisionMaker * DEC_MAK_PTR;

/*****************************************************************************/

class HeroAlwaysRight: public DecisionMaker {
	public:
										HeroAlwaysRight(HERO_PTR);
		virtual unsigned char 	applyPolicy (SKILL_OFFER_PTR offer);
		virtual void				LoopInitializer ();
};

/*****************************************************************************/

class HeroAlwaysLeft: public DecisionMaker {
	public:
										HeroAlwaysLeft(HERO_PTR);
		virtual unsigned char 	applyPolicy (SKILL_OFFER_PTR offer);
		virtual void				LoopInitializer ();
};

/*****************************************************************************/

class HeroAlwaysLeftThenPref: public DecisionMaker {
	public:
										HeroAlwaysLeftThenPref(HERO_PTR, unsigned char *);
		virtual unsigned char 	applyPolicy (SKILL_OFFER_PTR offer);
		virtual void				LoopInitializer ();

	private: 
		int				Preference[DIFF_SEC_SKILLS];
};

/*****************************************************************************/

class HeroSeekPrefOtherwiseUpgrade: public DecisionMaker {
	public:
										HeroSeekPrefOtherwiseUpgrade(HERO_PTR, unsigned char *);
		unsigned char				get_number_of_unwanted_skills () const;
		void							printPreferences ();
		virtual unsigned char 	applyPolicy (SKILL_OFFER_PTR offer);
		virtual void				LoopInitializer ();

	private:
		bool							qualifies_on_top (const unsigned char) const;
		int							Preference[DIFF_SEC_SKILLS];
		unsigned char				unwanted;
};

/*****************************************************************************/

enum Printing_Methods { DEPRECATED = 0, NEAT, FANCY };
#define MAX_PRINTING_METHOD FANCY

/*****************************************************************************/

class MathToolbox {
	public:
								MathToolbox ();
								~MathToolbox ();
		unsigned char		getPrintingMethod () const;
		bool					setPrintingMethod (const unsigned char);
		void					printStatistics (HERO_PTR, const clock_t, const clock_t);
		void					printStatistics_THREADS (HERO_PTR, TIME_DIFF *);
		//void					printStatistics_THREADS (HERO_PTR);
		bool 					incrementOccurrences (const unsigned char);
		bool 					setOccurrences (const unsigned char, const long unsigned int);
		long unsigned int getOccurrences (const unsigned char) const;
		void 					setNumberOfEpisodes (const long unsigned int);
		long unsigned int getNumberOfEpisodes () const;
		void 					setSeed (const unsigned int);
		unsigned int 		getSeed () const;
		void 					initializeOccurrences ();
		void 					setProbabilities ();
		void					smoothResults (HERO_PTR);


	private:
		void					printStatisticsDeprecated (HERO_PTR, const clock_t, const clock_t);
		void					printStatisticsNeat (HERO_PTR, const clock_t, const clock_t);
		void					printStatisticsFancy (HERO_PTR, const clock_t, const clock_t);
		void					printStatisticsDeprecated_THREADS (HERO_PTR, TIME_DIFF *);
		void					printStatisticsNeat_THREADS (HERO_PTR, TIME_DIFF *);
		void					printStatisticsFancy_THREADS (HERO_PTR, TIME_DIFF *);
		//void					printStatisticsDeprecated_THREADS (HERO_PTR);
		//void					printStatisticsNeat_THREADS (HERO_PTR);
		//void					printStatisticsFancy_THREADS (HERO_PTR);
		// Private variables.
		unsigned char		printing_method;
		long unsigned int occurrences [DIFF_SEC_SKILLS];
		long unsigned int episodes;
		double 				probabilities [DIFF_SEC_SKILLS];
		unsigned int 		seed;
};

typedef MathToolbox MT;
typedef MT * MT_PTR;

/*****************************************************************************/

class MonteCarlo  {
	public:
							MonteCarlo (HERO_PTR, DEC_MAK_PTR);
		void 				run (MT_PTR);
		void 				run_once (MT_PTR, const int *);

	private:
		HERO_PTR			myHero;
		DEC_MAK_PTR		myDecisionMaker;
};

typedef MonteCarlo MONTE_CARLO;
typedef MONTE_CARLO * MONTE_CARLO_PTR;

/*****************************************************************************/

#endif
