#ifndef __THREADS_DEFS_H__
#define __THREADS_DEFS_H__

#include <pthread.h>
#include "mighty_classes.h" // Necessary for MonteCarlo pointers...
#include "queue.h"			// Necessary for queues.

// Do not use more than 17 on the following.
#define NUM_THREADS 17	// Should be at least 2. Otherwise undefined behaviour.
#define WORKER_THREADS (NUM_THREADS - 1)	// id \in {0,1,2,...,NUM_THREADS-2}.
#define GENERATOR_ID (NUM_THREADS - 1)		// id: NUM_THREADS - 1.
// 44 elements are enough.
#define RND_NUMS_SIZE 44
#define EPISODES_FLUSH_CONSTANT 2000
// This should be the right value for 8 worker-threads.
// You can generate about 17 sequences per episode,
// so with 8 threads flush them in queues by pairs
// and time should be enough.
// I don't use that constant anywhere though ...
#define SEQS_TO_FLUSH_PER_ITERATION 2
#define PREPROCESSING_PERCENT 1.0

typedef struct {
	// Queue related variables.
	pthread_mutex_t **	queue_mutex_array;
	pthread_cond_t **	queue_cond_array;
	QUEUE_PTR *			queue_array;
	bool **				waiting_array;

	int					id;
	
	int					items_per_chunk;
} GENERATOR_DATA;

typedef struct {
	// Queue related stuff.
	pthread_mutex_t *	queue_mutex;
	pthread_cond_t *	queue_cond;
	QUEUE_PTR			queue;
	bool *				waiting;

	int					id;

	pthread_mutex_t *	total_episodes_mutex;	// See below for necessity.
	// The following variable is placed so that I can print
	// on screen a percent regarding the overall state of threads.
	// Note that only ONE thread should print at a time.
	long unsigned int * total_episodes_atm;

	long unsigned int	total_episodes_computed_by_threads;
	long unsigned int	episodes_to_compute;

	MONTE_CARLO_PTR		myMC;
	MT_PTR				myMT;
} WORKER_DATA;

typedef pthread_mutex_t * PTHREAD_MUTEX_T_PTR;
typedef pthread_cond_t * PTHREAD_COND_T_PTR;
typedef bool * BOOL_PTR;
typedef unsigned char * UNSIGNED_CHAR_PTR;	// for generator/worker functions.

#endif
